"""

:Author: Franziska Koehn
:Created: 2015/01/13

This module contains all classes relating to the widgets for creating a query.

"""

import gtk

class QueryView(gtk.VBox):

    def __init__(self, *args, **kwargs):
        super(QueryView, self).__init__(*args, **kwargs)

        import gobject
        gobject.signal_new( "rest-api",
                            QueryView,
                            gobject.SIGNAL_RUN_FIRST,
                            gobject.TYPE_NONE,
                            (bool,str))

        # root-type

        hBox_root_type = gtk.HBox()
        self.pack_start(hBox_root_type, False, True, 0)

        label_root_type = gtk.Label()
        label_root_type.set_text("Root-Type: ")
        hBox_root_type.pack_start(label_root_type, False, True, 0)

        def changed_cb(combobox):
            import xsa.datatypereader as type_reader
            import xsa.errors as xsa_errors
            selected =  combobox.get_active_text()
            try:
                type_reader.get_rest(selected)
            except xsa_errors.NoRestApiError as e:
                self.emit("rest-api", False, e)
            else:
                self.emit("rest-api", True, "")

            self.treeView_search.set_root_type(selected)
            self.treeView_fields.show_fields(selected)

        self.combobox = gtk.combo_box_new_text()

        import xsa.datatypereader as type_reader
        for t in type_reader.get_data_types():
            self.combobox.append_text(t)

        self.combobox.connect('changed', changed_cb)
        hBox_root_type.pack_start(self.combobox, True, True,0)

        # Sash

        vpaned_chart = gtk.VPaned()
        vpaned_chart.set_position(250)
        self.pack_start(vpaned_chart, True, True, 0)

        # query-tree
        sw_search = gtk.ScrolledWindow()
        sw_search.set_shadow_type(gtk.SHADOW_ETCHED_IN)
        sw_search.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        vpaned_chart.add1(sw_search)

        self.treeView_search = TreeViewQuery()
        sw_search.add(self.treeView_search)

        #result fields
        sw_fields = gtk.ScrolledWindow()
        sw_fields.set_shadow_type(gtk.SHADOW_ETCHED_IN)
        sw_fields.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        vpaned_chart.add2(sw_fields)

        self.treeView_fields = TreeViewResultFields()
        sw_fields.add(self.treeView_fields)

        # setting start state
        self.combobox.set_active(0)
        self.treeView_search.set_root_type(self.combobox.get_active_text())

    def get_query(self):
        return self.treeView_search.get_query(self.combobox.get_active_text())

    def get_root_type(self):
        return self.combobox.get_active_text()

    def get_fields(self):
        return self.treeView_fields.get_selected_fields()

    def set_data(self, root_field = None, query = None, field_labels = None):
        def get_index():
            return list(e[0] for e in self.combobox.get_model()).index(root_field)
        self.combobox.set_active(get_index())
        self.treeView_search.set_query(root_field, query)
        self.treeView_fields.set_fields(root_field, field_labels)

class TreeViewResultFields(gtk.TreeView):

    def __init__(self):
        super(TreeViewResultFields, self).__init__()
        self.create_columns()
        self.store = gtk.ListStore(bool, str, bool) # is_selected, label, is_required
        self.set_model(self.store)

    def create_columns(self):

        def callback_toggled(cellrenderertoggle, path_string, col, *_):
            it = self.store.get_iter_from_string(path_string)
            is_active = cellrenderertoggle.get_active()
            if not self.store.get_value(it, 2):
                self.store.set(it, col, not is_active)

        renderer = gtk.CellRendererToggle()
        renderer.set_property('activatable', True)
        renderer.connect("toggled", callback_toggled, 0)
        column = gtk.TreeViewColumn("", renderer)
        column.add_attribute(renderer, "active", 0)
        column.set_sort_column_id(0)
        column.set_resizable(False)
        self.append_column(column)

        column = gtk.TreeViewColumn("Requested Result-Fields", gtk.CellRendererText(), text=1)
        column.set_sort_column_id(1)
        column.set_resizable(True)
        self.append_column(column)

    def show_fields(self, root_type):
        import xsa.datatypereader as type_reader
        self.store.clear()
        fields = type_reader.get_fields_required(root_type)
        for f in fields:
            self.store.append( [ f[1],  # is_selected
                                f[0],   # label
                                f[1]    # is_required
                                ])

    def set_fields(self, root_type, fields):
        self.show_fields(root_type)
        for row in self.store:
            for field in fields:
                if row[1] == field:
                    row[0] = True
                    break

    def get_selected_fields(self):
        results = []
        for r in self.store:
            if r[0]: # if selected
                results.append(r[1])
        return results

class TreeViewQuery(gtk.TreeView):

    inital_value = "..."
    tooltip = 'Delete by using right click'
    root_type = ''
    methods = [] #["AND", "OR"]
    types = [] #["TYPE", "TE", "TI"]
    operators = [] #["LIKE", ">", "<", "<=", ">=", "="]


    def __init__(self):
        super(TreeViewQuery, self).__init__()

        self.create_type_list()
        self.create_method_list()
        self.create_operator_list()

        self.set_rules_hint(True)
        self.create_columns()
        self.set_headers_visible(True)
        self.create_model()
        self.set_model(self.store)

        def on_treeview_button_press_event(treeview, event):
            if event.button == 3:
                x = int(event.x)
                y = int(event.y)
                path_info = treeview.get_path_at_pos(x, y)
                if path_info is not None:
                    path, col, cellx, celly = path_info
                    iter = self.store.get_iter(path)
                    if not path == (0,):
                        md = gtk.MessageDialog(type=gtk.MESSAGE_WARNING,
                                               buttons=gtk.BUTTONS_YES_NO,
                                               message_format="Remove the element and its children?"
                                               )
                        response = md.run()
                        md.destroy()
                        if response == gtk.RESPONSE_YES:
                            self.store.remove(iter)

        self.connect('button_press_event', on_treeview_button_press_event)

        self.set_tooltip_column(4)

    def create_columns(self):

        def root_changed_callback(combo, path_string, new_iter, col, del_col, *_):

            it = self.store.get_iter_from_string(path_string)
            parent = self.store.iter_parent(it)
            new_value = combo.get_property("model")[new_iter][0]
            old_value = self.store.get_value(it, col)

            if new_value == self.inital_value or new_value == old_value:
                return
            elif old_value == self.inital_value:
                self.store.set(it, 0, new_value)
                if parent is not None:
                    self.store.append(parent,  (self.inital_value, '', '', False, self.tooltip))
            else:
                if new_value in self.types and old_value in self.methods:
                    md = gtk.MessageDialog(type=gtk.MESSAGE_WARNING,
                                           buttons=gtk.BUTTONS_YES_NO,
                                           message_format="The children will be destroyed!Continue?"
                                           )
                    response = md.run()
                    md.destroy()
                    if response == gtk.RESPONSE_NO:
                        return
                    else:
                        while not self.store.iter_children(it) is None:
                            child = self.store.iter_children(it)
                            self.store.remove(child)

                self.store.set(it, 0, new_value)


            if new_value in self.types:
                self.store.set(it, 3, True)
            else:
                self.store.set(it, 3, False)
                for c in del_col:
                    self.store.set(it, c, '')

                child_count = self.store.iter_n_children(it)
                if child_count == 0:
                    self.store.append(it,  (self.inital_value, '', '', False, self.tooltip))
                else:
                    last_child = self.store.iter_nth_child(it, child_count-1)
                    child_value = self.store.get_value(last_child, 0)
                    if child_value != self.inital_value:
                        self.store.append(it,  (self.inital_value, '', '', False, self.tooltip))


        def cell_changed_callback(combo, path_string, new_iter, col, *_):
            it = self.store.get_iter_from_string(path_string)
            text = combo.get_property("model")[new_iter][0]
            self.store.set(it, col, str(text))

        def cell_edited_callback(cellrenderertext, path_string, new_text, col, *_):
            it = self.store.get_iter_from_string(path_string)
            self.store.set(it, col, new_text)

        rendererCombo = gtk.CellRendererCombo()
        rendererCombo.set_property('editable', True)
        rendererCombo.set_property ("model", self.create_combo_list_type());
        rendererCombo.set_property ("text-column", 0);
        rendererCombo.connect('changed', root_changed_callback, 0, [1,2])

        column = gtk.TreeViewColumn("Query", rendererCombo, text=0)
        column.set_sort_column_id(0)
        column.set_min_width(100)
        column.set_resizable(True)
        self.append_column(column)

        rendererCombo = gtk.CellRendererCombo()
        rendererCombo.set_property('editable', True)
        rendererCombo.set_property ("model", self.create_combo_list_operator())
        rendererCombo.set_property ("text-column", 0);
        rendererCombo.connect('changed', cell_changed_callback, 1)

        column = gtk.TreeViewColumn("Operator", rendererCombo, text=1, editable=3)
        column.set_sort_column_id(1)
        column.set_min_width(70)
        self.append_column(column)

        rendererText = gtk.CellRendererText()
        rendererText.set_property('editable', True)
        rendererText.connect('edited', cell_edited_callback, 2)

        column = gtk.TreeViewColumn("Value", rendererText, text=2, editable=3)
        column.set_sort_column_id(2)
        column.set_min_width(100)
        column.set_resizable(True)

        self.append_column(column)

    @staticmethod
    def create_combo_list(labels):
        combo_model = gtk.ListStore(str)
        for l in labels:
            combo_model.append((str(l),))
        return combo_model

    def set_root_type(self, root_type):
        self.root_type = root_type
        self.reset_model()
        self.create_type_list()
        self.get_column(0).get_cell_renderers()[0].set_property ("model", self.create_combo_list_type()); #TODO column number???

    def create_combo_list_type(self):
        return self.create_combo_list(self.methods+self.types)

    def create_combo_list_operator(self):
        return self.create_combo_list(self.operators)

    def reset_model(self):
        self.store.clear()
        self.store.append(None, (self.inital_value, '', '', False, self.tooltip))

    def create_model(self):
        self.store = gtk.TreeStore(str, str, str, bool, str)
        self.store.append(None, (self.inital_value, '', '', False, self.tooltip))

    def create_type_list(self):
        import xsa.datatypereader as type_reader
        self.types = type_reader.get_labels(self.root_type)

    def create_method_list(self):
        import xsa.queries as queries
        self.methods = queries.get_query_methods()

    def create_operator_list(self):
        import xsa.queries as queries
        self.operators = queries.get_operators()

    def set_query(self, root_type, query):
        import xsa.errors as xsa_errors
        import xsa.datatypereader as type_reader

        def add_next_to_store(values, parent):
            if type(values) is tuple or (len(values) == 3 and all(type(v) is unicode for v in values)):
                type_name = type_reader.get_field_label_by_field(root_type, values[0])
                self.store.append(parent,
                    (type_name,values[1],values[2],True,self.tooltip))
            elif type(values) is list:
                for v in values:
                    if v in self.methods:
                        parent = self.store.append(parent,
                                    (v, '', '', False, self.tooltip))
                    else:
                        add_next_to_store(v, parent)
            else:
                raise xsa_errors.CorruptedQueryError("Error while reading Query")

        self.store.clear()
        add_next_to_store(query[0], None)


    def get_query(self, type):
        import xsa.datatypereader as type_reader
        search_fields = type_reader.get_fields(type)

        def add_next_to_query(iter, result):
            value = self.store.get_value(iter, 0)
            constraint = []
            if value in self.methods:
                result.append(constraint)
                constraint.append(value)
            elif value in self.types:
                for field in search_fields:
                    if field['label'] == value:
                        value = field['field']
                        break
                operator = self.store.get_value(iter, 1)
                field_value = self.store.get_value(iter, 2)
                result.append( (value, operator, field_value) )
            child = self.store.iter_children(iter)
            if not child is None:
                add_next_to_query(child, constraint)
            next_ = self.store.iter_next(iter)
            if not next_ is None:
                add_next_to_query(next_, result)

        it = self.store.get_iter_first()
        query = []
        add_next_to_query(it, query)
        return query
